import logging
from collections.abc import MutableMapping
from typing import Callable, Optional

logger = logging.getLogger(__name__)


class ProcessRegistry(MutableMapping):
    """
    The process registry is basically a dictionary mapping from process_id to the callable implementation.
    It also allows registering aliases for process_ids.
    """

    def __init__(self, wrap_funcs: Optional[list] = None, *args, **kwargs):
        """wrap_funcs: list of decorators to apply to all registered processes."""
        self.store = dict()  # type: dict[str, Callable]
        self.aliases = dict()  # type: dict[str, str]

        if wrap_funcs is None:
            wrap_funcs = []
        self.wrap_funcs = wrap_funcs

        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key) -> Callable:
        t_key = self._keytransform(key)
        if t_key in self.store:
            return self.store[t_key]
        if t_key in self.aliases:
            original_key = self.aliases[t_key]
            if original_key in self.store:
                return self.store[original_key]
            else:
                del self.aliases[t_key]

        raise KeyError(f"Key {key} not found in process registry!")

    def __setitem__(self, key, value):
        t_key = self._keytransform(key)

        decorated_value = value
        for wrap_f in self.wrap_funcs:
            decorated_value = wrap_f(decorated_value)

        self.store[t_key] = decorated_value

    def __delitem__(self, key):
        t_key = self._keytransform(key)

        del self.store[t_key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def _keytransform(self, key):
        """Some processes are prefixed with an underscore to prevent clashes with built-in names.
        These need to be stripped before being put into the registry."""
        return key.strip("_")

    def add_alias(self, process_id: str, alias: str):
        """
        Method to allow adding aliases to processes.
        This can be useful for not-yet standardised processes, where an OpenEO client might use a different process_id than the backend.
        """

        if process_id not in self.store:
            raise ValueError(
                f"Could not add alias {alias} -> {process_id}, because process_id {process_id} was not found in the process registry."
            )

        # Add the alias to the self.aliases dict
        self.aliases[self._keytransform(alias)] = self._keytransform(process_id)
        logger.debug(f"Added alias {alias} -> {process_id} to process registry.")

    def add_wrap_func(self, wrap_func: Callable):
        self.wrap_funcs.append(wrap_func)

        # Wrap all existing processes retroactively
        for key, value in self.items():
            self[key] = wrap_func(value)



from __future__ import annotations

import datetime
import json
import logging
from enum import Enum
from re import match
from typing import Any, Optional, Union
from uuid import UUID, uuid4

import numpy as np
import pendulum
import pyproj
from geojson_pydantic import (
    Feature,
    FeatureCollection,
    GeometryCollection,
    MultiPolygon,
    Polygon,
)
from pydantic import BaseModel, Extra, Field, ValidationError, constr, validator
from shapely.geometry import Polygon

logger = logging.getLogger(__name__)

# TODO: Move this to a proper settings object for this repo and use in tests, possibly using pydantic settings: https://pydantic-docs.helpmanual.io/usage/settings/
DEFAULT_CRS = pyproj.CRS.from_user_input("EPSG:4326").to_wkt()

# This controls what is imported when calling `from pg_schema import *`, just a shortcut to import all types.
__all__ = [
    "ResultReference",
    "ParameterReference",
    "ProcessNode",
    "ProcessGraph",
    "PGEdgeType",
    "BoundingBox",
    "Year",
    "Date",
    "DateTime",
    "Duration",
    "Time",
    "TemporalInterval",
    "TemporalIntervals",
    "GeoJson",
    "JobId",
    "OutputFormat",
    "DEFAULT_CRS",
]


class ResultReference(BaseModel, extra=Extra.forbid):
    from_node: str
    node: ProcessNode


class ParameterReference(BaseModel, extra=Extra.forbid):
    from_parameter: str


class ProcessNode(BaseModel, arbitrary_types_allowed=True):
    process_id: constr(regex=r'^\w+$')
    namespace: Optional[Optional[str]] = None
    result: Optional[bool] = False
    description: Optional[Optional[str]] = None
    arguments: dict[
        str,
        Optional[
            Union[
                ResultReference,
                ParameterReference,
                ProcessGraph,
                BoundingBox,
                JobId,
                OutputFormat,
                Year,
                Date,
                DateTime,
                Duration,
                TemporalInterval,
                TemporalIntervals,
                # GeoJson, disable while https://github.com/developmentseed/geojson-pydantic/issues/92 is open
                Time,
                float,
                str,
                bool,
                list,
                dict,
            ]
        ],
    ]

    def __str__(self):
        return json.dumps(self.dict(), indent=4)


class ProcessGraph(BaseModel, extra=Extra.forbid):
    process_graph: dict[str, ProcessNode]
    uid: UUID = Field(default_factory=uuid4)


class PGEdgeType(str, Enum):
    ResultReference = "result_reference"
    Callback = "callback"


def parse_crs(v) -> pyproj.CRS:
    if v is None or v.strip() == "":
        return DEFAULT_CRS
    else:
        try:
            # Check that the crs can be parsed and store as WKT
            crs_obj = pyproj.CRS.from_user_input(v)
            return crs_obj.to_wkt()
        except pyproj.exceptions.CRSError as e:
            logger.error(f"Provided CRS {v} could not be parsed, defaulting to EPSG:4326")
            raise e


def crs_validator(field: str) -> classmethod:
    decorator = validator(field, allow_reuse=True, pre=True, always=True)
    validator_func = decorator(parse_crs)
    return validator_func


class BoundingBox(BaseModel, arbitrary_types_allowed=True):
    west: float
    east: float
    north: float
    south: float
    base: Optional[float]
    height: Optional[float]
    crs: Optional[str]

    # validators
    _parse_crs: classmethod = crs_validator('crs')

    @property
    def polygon(self) -> Polygon:
        """"""
        return Polygon(
            [
                (self.west, self.south),
                (self.west, self.north),
                (self.east, self.north),
                (self.east, self.south),
            ]
        )


class Date(BaseModel):
    __root__: datetime.datetime

    @validator("__root__", pre=True)
    def validate_time(cls, value: Any) -> Any:
        if (
            isinstance(value, str)
            and len(value) <= 11
            and match(r"[0-9]{4}[-/][0-9]{2}[-/][0-9]{2}T?", value)
        ):
            return pendulum.parse(value)
        raise ValidationError("Could not parse `Date` from input.")

    def to_numpy(self):
        raise NotImplementedError

    def __repr__(self):
        return self.__root__.__repr__()


class DateTime(BaseModel):
    __root__: datetime.datetime

    @validator("__root__", pre=True)
    def validate_time(cls, value: Any) -> Any:
        if isinstance(value, str) and match(
            r"[0-9]{4}-[0-9]{2}-[0-9]{2}T?[0-9]{2}:[0-9]{2}:?([0-9]{2})?Z?", value
        ):
            return pendulum.parse(value)
        raise ValidationError("Could not parse `DateTime` from input.")

    def to_numpy(self):
        return np.datetime64(self.__root__)

    def __repr__(self):
        return self.__root__.__repr__()


class Time(BaseModel):
    __root__: pendulum.Time

    @validator("__root__", pre=True)
    def validate_time(cls, value: Any) -> Any:
        if (
            isinstance(value, str)
            and len(value) >= 5
            and len(value) <= 9
            and match(r"[0-9]{2}:[0-9]{2}:?([0-9]{2})?Z?", value)
        ):
            return pendulum.parse(value).time()
        raise ValidationError("Could not parse `Time` from input.")

    def to_numpy(self):
        raise NotImplementedError

    def __repr__(self):
        return self.__root__.__repr__()


class Year(BaseModel):
    __root__: datetime.datetime

    @validator("__root__", pre=True)
    def validate_time(cls, value: Any) -> Any:
        if isinstance(value, str) and len(value) <= 4 and match(r"^\d{4}$", value):
            return pendulum.parse(value)
        raise ValidationError("Could not parse `Year` from input.")

    def to_numpy(self):
        return np.datetime64(self.__root__)

    def __repr__(self):
        return self.__root__.__repr__()


class Duration(BaseModel):
    __root__: pendulum.Duration

    @validator("__root__", pre=True)
    def validate_time(cls, value: Any) -> Any:
        if isinstance(value, str) and match(
            r"P[0-9]*Y?[0-9]*M?[0-9]*D?T?[0-9]*H?[0-9]*M?[0-9]*S?", value
        ):
            return pendulum.parse(value)
        raise ValidationError("Could not parse `Duration` from input.")

    def to_numpy(self):
        raise NotImplementedError

    def __repr__(self):
        return self.__root__.__repr__()


class TemporalInterval(BaseModel):
    __root__: list[Union[Year, Date, DateTime, Time]]

    @property
    def start(self):
        return self.__root__[0]

    @property
    def end(self):
        return self.__root__[1]

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]


class TemporalIntervals(BaseModel):
    __root__: list[TemporalInterval]

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item) -> TemporalInterval:
        return self.__root__[item]


GeoJson = Union[FeatureCollection, Feature, GeometryCollection, MultiPolygon, Polygon]
# The GeoJson spec (https://www.rfc-editor.org/rfc/rfc7946.html#ref-GJ2008) doesn't
# have a crs field anymore and recommends assuming it to be EPSG:4326, so we do the same.


class JobId(BaseModel):
    __root__: str = Field(
        regex=r"(eodc-jb-|jb-)[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}"
    )


class OutputFormat(BaseModel):
    __root__: str = Field(regex=r"(?i)(gtiff|geotiff|netcdf|json)")


ResultReference.update_forward_refs()
ProcessNode.update_forward_refs()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openeo_pg_parser_networkx']

package_data = \
{'': ['*']}

install_requires = \
['geojson-pydantic>=0.5.0,<0.6.0',
 'networkx[default]>=2.8.6,<3.0.0',
 'numpy>=1.24.1,<2.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'pyproj>=3.4.0,<4.0.0',
 'shapely>=1.8.4,<2.0.0']

setup_kwargs = {
    'name': 'openeo-pg-parser-networkx',
    'version': '2023.1.1',
    'description': 'Parse OpenEO process graphs from JSON to traversible Python objects.',
    'long_description': '# OpenEO Process Graph Parser (Python & networkx)\n[![codecov](https://codecov.io/github/Open-EO/openeo-pg-parser-networkx/branch/main/graph/badge.svg?token=KEAKFB8AFX)](https://codecov.io/github/Open-EO/openeo-pg-parser-networkx)\n\nPython package to parse OpenEO process graphs from raw JSON into fully traversible [`networkx`](https://github.com/networkx/networkx) graph objects.\nThis package is an evolution of the [openeo-pg-parser-python](https://github.com/Open-EO/openeo-pg-parser-python) package.\n\n## Installation\nThis package can be installed with pip:\n\n```\npip install openeo-pg-parser-networkx\n```\n\nCurrently Python versions 3.9 and 3.10 are supported.\n\n## Basic usage\n(An example notebook of using `openeo-pg-parser-networkx` together with a process implementation source like [`openeo-processes-dask`](https://github.com/Open-EO/openeo-processes-dask) can be found in `openeo-pg-parser-networkx/examples/01_minibackend_demo.ipynb`.)\n\n**Parse a JSON OpenEO process graph:**\n\n```\nfrom openeo_pg_parser_networkx.graph import OpenEOProcessGraph\n\nNDVI_GRAPH_PATH = "../tests/data/graphs/apply.json"\n\nparsed_graph = OpenEOProcessGraph.from_file(NDVI_GRAPH_PATH)\n```\n\n```\n> Deserialised process graph into nested structure\n> Walking node root-fd8ae3b4-8cb8-46c8-a5cd-c8ee552d1945\n> Walking node apply2-fd8ae3b4-8cb8-46c8-a5cd-c8ee552d1945\n> Walking node multiply1-f8644201-32a8-4283-8814-a577c4e28226\n> Walking node apply1-fd8ae3b4-8cb8-46c8-a5cd-c8ee552d1945\n> Walking node ndvi1-06a8d8af-296a-4960-a1cb-06dcd251b6bb\n> Walking node loadcollection1-fd8ae3b4-8cb8-46c8-a5cd-c8ee552d1945\n```\n\n**Plot it:**\n\n```\nparsed_graph.plot()\n```\n\n![example process graph](./examples/images/apply_ndvi.png)\n\nTo execute a process graph, `OpenEOProcessGraph` needs to know which Python code to call for each of the nodes in the graph. This information is provided by a "process registry", which is basically a dictionary that maps each `process_id` to its actual Python implementation as a `Callable`.\n\n**Register process implementations to a "process registry":**\n\n```\nfrom openeo_pg_parser_networkx import ProcessRegistry\nprocess_registry = ProcessRegistry()\n\nfrom openeo_processes_dask.process_implementations import apply, ndvi, multiply, load_collection, save_result\n\nprocess_registry["apply"] =  apply\nprocess_registry["ndvi"] =  ndvi\nprocess_registry["multiply"] =  multiply\nprocess_registry["load_collection"] =  load_collection\nprocess_registry["save_result"] =  save_result\n```\n\n**Build an executable callable from the process graph:**\n\n```\npg_callable = parsed_graph.to_callable(process_registry=process_registry)\n```\n\n**Execute that callable like a normal Python function:**\n\n```\npg_callable\n```\n\n```\n> Running process load_collection\n> Running process apply\n> ...\n```\n\n## Development environment\n`openeo-pg-parser-networkx` requires poetry `>1.2`, see their [docs](https://python-poetry.org/docs/#installation) for installation instructions.\n\nTo setup the python venv and install this project into it run:\n```\npoetry install\n```\n\nTo add a new core dependency run:\n```\npoetry add some_new_dependency\n```\n\nTo add a new development dependency run:\n```\npoetry add some_new_dependency --group dev\n```\n\nTo run the test suite run:\n```\npoetry run python -m pytest\n```\n\nNote that you can also use the virtual environment that\'s generated by poetry as the kernel for the ipynb notebooks.\n\n### Pre-commit hooks\nThis repo makes use of [pre-commit](https://pre-commit.com/) hooks to enforce linting & a few sanity checks.\nIn a fresh development setup, install the hooks using `poetry run pre-commit install`.\nThese will then automatically be checked against your changes before making the commit.\n',
    'author': 'Lukas Weidenholzer',
    'author_email': 'lukas.weidenholzer@eodc.eu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Open-EO/openeo-pg-parser-networkx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



# This is forked from openeo-python-client, because I needed to have this unflatten subgraphs

from typing import Any

import pydantic

from openeo_pg_parser_networkx.pg_schema import ParameterReference, ResultReference


def parse_nested_parameter(parameter: Any):
    try:
        return ResultReference.parse_obj(parameter)
    except pydantic.error_wrappers.ValidationError:
        pass
    except TypeError:
        pass

    try:
        return ParameterReference.parse_obj(parameter)
    except pydantic.error_wrappers.ValidationError:
        pass
    except TypeError:
        pass

    return parameter


def find_result_node(flat_graph: dict) -> tuple[str, dict]:
    """
    Find result node in flat graph

    :return: tuple with node id (str) and node dictionary of the result node.
    """
    result_nodes = [
        (key, node) for (key, node) in flat_graph.items() if node.get("result")
    ]

    if len(result_nodes) == 1:
        return result_nodes[0]
    elif len(result_nodes) == 0:
        raise Exception("Found no result node in flat process graph")
    else:
        keys = [k for (k, n) in result_nodes]
        raise Exception(
            "Found multiple result nodes in flat process graph: {keys!r}".format(
                keys=keys
            )
        )


class ProcessGraphUnflattener:
    """
    Base class to process a flat graph representation of a process graph
    and unflatten it by resolving the "from_node" references.
    Subclassing and overriding certain methods allows to build a desired unflattened graph structure.
    """

    # Sentinel object for flagging a node "under construction" and detect graph cycles.
    _UNDER_CONSTRUCTION = object()

    def __init__(self, flat_graph: dict):
        self._flat_graph = flat_graph
        self._nodes = {}

    @classmethod
    def unflatten(cls, flat_graph: dict, **kwargs):
        """Class method helper to unflatten given flat process graph"""
        return cls(flat_graph=flat_graph, **kwargs).process()

    def process(self):
        """Process the flat process graph: unflatten it."""
        result_key, result_node = find_result_node(flat_graph=self._flat_graph)
        return self.get_node(result_key)

    def get_node(self, key: str) -> Any:
        """Get processed node by node key."""
        if key not in self._nodes:
            self._nodes[key] = self._UNDER_CONSTRUCTION
            node = self._process_node(self._flat_graph[key])
            self._nodes[key] = node
        elif self._nodes[key] is self._UNDER_CONSTRUCTION:
            raise Exception("Cycle in process graph")
        return self._nodes[key]

    def _process_node(self, node: dict) -> Any:
        """
        Overridable: generate process graph node from flat_graph data.
        """
        # Default implementation: basic validation/whitelisting, and only traverse arguments
        return dict(
            process_id=node["process_id"],
            arguments=self._process_value(value=node["arguments"]),
            **{k: node[k] for k in ["namespace", "description", "result"] if k in node},
        )

    def _process_from_node(self, key: str, node: dict) -> Any:
        """
        Overridable: generate a node from a flat_graph "from_node" reference
        """
        # Default/original implementation: keep "from_node" key and add resolved node under "node" key.
        # TODO: just return `self.get_node(key=key)`
        return {"from_node": key, "node": self.get_node(key=key)}

    def _process_from_parameter(self, name: str) -> Any:
        """
        Overridable: generate a node from a flat_graph "from_parameter" reference
        """
        # Default implementation:
        return {"from_parameter": name}

    def _process_child_graph(self, node_name: str, child_graph: dict):
        return {"process_graph": {node_name: child_graph}}

    def _resolve_from_node(self, key: str) -> dict:
        if key not in self._flat_graph:
            raise Exception(f"from_node reference {key!r} not found in process graph")
        return self._flat_graph[key]

    def _process_value(self, value) -> Any:
        if isinstance(value, dict):
            if "from_node" in value:
                key = value["from_node"]
                node = self._resolve_from_node(key=key)
                return self._process_from_node(key=key, node=node)
            elif "from_parameter" in value:
                name = value["from_parameter"]
                return self._process_from_parameter(name=name)
            elif "process_graph" in value:
                result_node_id, _ = find_result_node(value["process_graph"])
                return self._process_child_graph(
                    node_name=result_node_id,
                    child_graph=ProcessGraphUnflattener.unflatten(value["process_graph"]),
                )
            else:
                return {k: self._process_value(v) for (k, v) in value.items()}
        elif isinstance(value, (list, tuple)):
            return [self._process_value(v) for v in value]
        else:
            return value
